@@ -5,7 +5,8 @@
 <https://atlas.ripe.net/> probes using the UDM (User Defined
 Measurements) creation API.
 
-Authorization key is expected in $HOME/.atlas/auth or have to be
+Authorization key is expected in $HOME/.atlas/auth
+in the environment variable ATLASAUTH or have to be
 provided in the constructor's arguments.
 
 Stéphane Bortzmeyer <stephane+frama@bortzmeyer.org>
@@ -13,7 +14,7 @@ Stéphane Bortzmeyer <stephane+frama@bortzmeyer.org>
 """
 
 # WARNING: if you modify it here, also change setup.py https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version
-VERSION = '1.1.8'
+VERSION = '2.0.0'
 
 import os
 import json
@@ -27,6 +28,7 @@ import string
 
 authfile = "%s/.atlas/auth" % os.environ['HOME']
 base_url = "https://atlas.ripe.net/api/v2/measurements"
+base_url_probes = "https://atlas.ripe.net/api/v2/probes"
 
 # The following parameters are currently not settable. Anyway, be
 # careful when changing these, you may get inconsistent results if you
@@ -52,7 +54,10 @@ class AuthFileEmpty(Exception):
     pass
 
 class RequestSubmissionError(Exception):
-    pass
+    def __init__(self, http_status, reason, body):
+        self.status = http_status
+        self.reason = reason
+        self.body = body
 
 class FieldsQueryError(Exception):
     pass
@@ -63,6 +68,12 @@ class MeasurementNotFound(Exception):
 class MeasurementAccessError(Exception):
     pass
 
+class ProbeNotFound(Exception):
+    pass
+
+class ProbeAccessError(Exception):
+    pass
+
 class ResultError(Exception):
     pass
 
@@ -76,6 +87,16 @@ class InternalError(Exception):
 class WrongAssumption(Exception):
     pass
 
+# Utilities
+def format_error(error):
+    msg = ""
+    if error.status is not None:
+        msg += "HTTP status %s " % error.status
+    msg += error.reason
+    if error.body is not None and error.body != "":
+        msg += " %s" % error.body
+    return msg
+
 class Config:
     def __init__(self):
         # Default values
@@ -93,8 +114,12 @@ class Config:
         self.machine_readable = False
         self.measurement_id = None
         self.display_probes = False
+        self.display_probe_asns = False
+        self.cache_probes = None
         self.ipv4 = False
         self.private = False
+        self.resolve_on_probe = False
+        self.tags = None # Tags we send to tag the measurement, not tags of the probes
         self.port = 80
         self.size = 64
         self.spread = None
@@ -109,6 +134,8 @@ class Config:
         --verbose or -v : makes the program more talkative
         --help or -h : this message
         --displayprobes or -o : display the probes numbers (WARNING: big lists)
+        --displayprobeasns : display the (unique) probe ASNumbers (currently only for blaeu-resolve)
+        --cache-probes=<file.json> : cache probe data
         --country=2LETTERSCODE or -c 2LETTERSCODE : limits the measurements to one country (default is world-wide)
         --area=AREACODE or -a AREACODE : limits the measurements to one area such as North-Central (default is world-wide)
         --asn=ASnumber or -n ASnumber : limits the measurements to one AS (default is all ASes)
@@ -123,8 +150,10 @@ class Config:
         --port=N or -t N : destination port for TCP (default is %s)
         --size=N or -z N : number of bytes in the packet (default is %s bytes)
         --ipv4 or -4 : uses IPv4 (default is IPv6, except if the parameter or option is an IP address, then it is automatically found)
+        --tags TAGS : tag the measurement (no relationship with probes tags) (a comma-separated list)
         --spread or -w : spreads the tests (add a delay before the tests)
         --private : makes the measurement private
+        --resolve-on-probe : resolve names with the probe's DNS resolver
         --machinereadable or -b : machine-readable output, to be consumed by tools like grep or cut
         """ % (self.requested, int(self.percentage_required*100), self.port, self.size), file=sys.stderr)
 
@@ -133,11 +162,12 @@ class Config:
             usage = self.usage
         try:
             optlist, args = getopt.getopt (sys.argv[1:],
-                                           "4a:bc:e:f:g:hi:m:n:op:r:s:t:vw:z:" + shortOptsSpecific,
-                                           ["requested=", "country=", "area=", "asn=", "prefix=", "probes=",
+                                           "4a:bc:e:f:g:hi:m:n:op:r:s:t:uvw:z:" + shortOptsSpecific,
+                                           ["requested=", "country=", "area=", "asn=", "prefix=", "cache-probes=", "probes=",
                                             "port=", "percentage=", "include=", "exclude=", "version",
-                                            "measurement-ID=", "old_measurement=", "displayprobes", "size=",
-                                            "ipv4", "private", "machinereadable", "spread=", "verbose", "help"] +
+                                            "measurement-ID=", "old_measurement=", "displayprobes", "displayprobeasns", "size=",
+                                            "ipv4", "private", "resolve-on-probe",
+                                            "machinereadable", "spread=", "verbose", "tags=", "help"] +
                                            longOptsSpecific)
             for option, value in optlist:
                 if option == "--country" or option == "-c":
@@ -148,6 +178,8 @@ class Config:
                     self.asn = value
                 elif option == "--prefix" or option == "-f":
                     self.prefix = value
+                elif option == "--cache-probes":
+                    self.cache_probes = value
                 elif option == "--probes" or option == "-s":
                     self.probes = value # Splitting (and syntax checking...) delegated to Atlas
                 elif option == "--percentage" or option == "-p":
@@ -167,17 +199,23 @@ class Config:
                     self.ipv4 = True
                 elif option == "--private":
                     self.private = True
+                elif option == "--resolve-on-probe" or option == "-u":
+                    self.resolve_on_probe = True
                 elif option == "--size" or option == "-z":
                     self.size = int(value)
                 elif option == "--spread" or option == "-w":
                     self.spread = int(value)
                 elif option == "--displayprobes" or option == "-o":
                     self.display_probes = True
+                elif option == "--displayprobeasns":
+                    self.display_probe_asns = True
                 elif option == "--exclude" or option == "-e":
                     self.exclude = value.split(",")
                 elif option == "--include" or option == "-i":
                     # TODO allows to specify stable probes https://labs.ripe.net/Members/chris_amin/new-ripe-atlas-probe-stability-system-tags
                     self.include = value.split(",")
+                elif option == "--tags":
+                    self.tags = value.split(",")
                 elif option == "--machinereadable" or option == "-b":
                     self.machine_readable = True
                 elif option == "--help" or option == "-h":
@@ -279,8 +317,12 @@ class Config:
             data["definitions"][0]['af'] = 6 
         if self.private:
             data["definitions"][0]['is_public'] = False
+        if self.resolve_on_probe:
+            data["definitions"][0]['resolve_on_probe'] = True
         if self.size is not None:
-            data["definitions"][0]['size'] = self.size    
+            data["definitions"][0]['size'] = self.size
+        if self.tags is not None:
+            data["definitions"][0]["tags"] = self.tags
         if self.spread is not None:
             data["definitions"][0]['spread'] = self.spread
         data["probes"][0]["tags"] = {}
@@ -324,18 +366,23 @@ class Measurement():
         """
 
         if data is None and id is None:
-            raise RequestSubmissionError("No data and no measurement ID")
+            raise RequestSubmissionError(None, "No data and no measurement ID", None)
         
         # TODO: when creating a dummy measurement, a key may not be necessary if the measurement is public
         if not key:
-            if not os.path.exists(authfile):
-                raise AuthFileNotFound("Authentication file %s not found" % authfile)
-            auth = open(authfile)
-            key = auth.readline()
-            if key is None or key == "":
-                raise AuthFileEmpty("Authentication file %s empty or missing a end-of-line at the end" % authfile)
-            key = key.rstrip('\n')
-            auth.close()
+            if os.environ.get("ATLASAUTH"):
+                # use envvar ATLASAUTH for the key
+                key = os.environ.get("ATLASAUTH")
+            else:
+                # use file for key if envvar ATLASAUTH isn't set
+                if not os.path.exists(authfile):
+                    raise AuthFileNotFound("Authentication file %s not found" % authfile)
+                auth = open(authfile)
+                key = auth.readline()
+                if key is None or key == "":
+                    raise AuthFileEmpty("Authentication file %s empty or missing a end-of-line at the end" % authfile)
+                key = key.rstrip('\n')
+                auth.close()
 
         self.url = base_url + "/?key=%s" % key
         self.url_probes = base_url + "/%s/?fields=probes,status" + "&key=%s" % key
@@ -358,12 +405,9 @@ class Measurement():
                 self.id = results["measurements"][0]
                 conn.close()
             except urllib.error.HTTPError as e:
-                raise RequestSubmissionError("Status %s, reason \"%s : %s\"" % \
-                                             (e.code, e.reason, e.read()))
+                raise RequestSubmissionError(e.code, e.reason, e.read().decode())
             except urllib.error.URLError as e:
-                raise RequestSubmissionError("Reason \"%s\"" % \
-                                             (e.reason))
-
+                raise RequestSubmissionError(None, e.reason, None)
 
             self.gen = random.Random()
             self.time = time.gmtime()
@@ -521,3 +565,95 @@ class Measurement():
             except urllib.error.URLError as e:
                 raise ResultError(e.reason)
         return result_data
+
+class Probe():
+    """ An Atlas probe, identified by its ID (such as #1010569) in the field "id" """
+    __probe_attributes = {
+        "id": "id",
+        "status": "status.name",
+        "address_v4": "address_v4",
+        "address_v6": "address_v6",
+        "asn_v4": "asn_v4",
+        "asn_v6": "asn_v6",
+        "country_code": "country_code",
+        "is_public": "is_public",
+        "prefix_v4": "prefix_v4",
+        "prefix_v6": "prefix_v6",
+        "tags": "tags",
+        "description": "description",
+        "geometry": "geometry",
+        "first_connected": "first_connected",
+        "last_connected": "last_connected",
+    }
+
+    def __init__(self, id, fetch=True):
+        if not fetch:
+            self.id = id
+            return
+        self.url = base_url_probes + "/%s" % id
+        try:
+            conn = urllib.request.urlopen(JsonRequest(self.url))
+        except urllib.error.HTTPError as e:
+            if e.code == 404:
+                raise ProbeNotFound
+            else:
+                raise ProbeAccessError("HTTP %s, %s %s" % (e.code, e.reason, e.read()))
+        except urllib.error.URLError as e:
+            raise ProbeAccessError("Reason \"%s\"" % \
+                                         (e.reason))
+        probe_data = json.loads(conn.read().decode('utf-8'))
+        def nested_get(d, keys):
+            k = keys.pop(0)
+            d = d[k]
+            return nested_get(d, keys) if keys else d
+        for key, lookup_key in self.__probe_attributes.items():
+            setattr(self, key, nested_get(probe_data, lookup_key.split(".")))
+
+    def __str__(self):
+        return "Probe #%s, %s" % (self.id, self.description)
+
+    def __repr__(self):
+        return "Probe #%s, %s" % (self.id, self.description)
+
+    def __eq__(self, other):
+        return self.__dict__ == other.__dict__
+
+    @classmethod
+    def from_dict(cls, data):
+        """Creates a Probe object from a dictionary (not a JSON string)"""
+        if not all(k in data for k in cls.__probe_attributes):
+            return
+        probe = cls(data["id"], fetch=False)
+        for key, value in data.items():
+            setattr(probe, key, value)
+        return probe
+
+class ProbeCache():
+    """ A cache for probe data, to avoid querying the Atlas servers each time. """
+    __cache__ = {}
+
+    def __init__(self, filename):
+        self.filename = filename
+        self.data = {} if not os.path.exists(filename) else \
+            json.load(open(filename), object_hook=lambda o: Probe.from_dict(o) or o)
+
+    def __str__(self):
+        return "ProbeCache %s, %s probes" % (self.filename, len(self.data))
+
+    def __repr__(self):
+        return "ProbeCache %s, %s probes" % (self.filename, len(self.data))
+
+    def get(self, probe):
+        return self.data.get(str(probe.id))
+
+    def put(self, probe):
+        prev_data = self.data.copy()
+        self.data[str(probe.id)] = probe
+        if self.data != prev_data:
+            json.dump(self.data, open(self.filename, "w"), default=lambda o: o.__dict__)
+        return probe
+
+    @classmethod
+    def cache_probe_id(cls, filename, id):
+        cache = cls.__cache__.setdefault(filename, cls(filename))
+        return cache.get(Probe(id)) or cache.put(Probe(id))
